from meissner.logger import *

class Filter:
    def test(self, payload):
        logging.warning('Using dummy filter ', colored_command(self.__class__.__name__))
        return payload

from .urlencode import URLEncoder



from meissner.logger import *
import meissner

from smartbytes import *
import threading, time


class Meissner:
    '''
    arg dictionary: must be a list of smartbytes
    '''
    def __init__(self, harness, dictionary, filters = None, engine = None):
        self.harness = harness
        self.dictionary = dictionary
        self.filters = filters or list()
        self.engine = engine or meissner.engines.Chrome


    '''
    test a payload and return whether or not it popped an alert dialog
    '''
    def _format(self, payload):
        return self._filter(self._harness(payload))


    '''
    pass payload through all configured filters
    '''
    def _filter(self, payload):
        for filter in self.filters:
            payload = smartbytes(filter.filter(payload))

        return payload

    '''
    test xss payload through harness post-filter
    '''
    def _harness(self, payload):
        return smartbytes(self.harness.test(payload))

    '''
    finish job
    '''
    def stop(self):
        logging.debug('Waiting for mutator to clean up...')
        self.mutator.running = False
        self.mutator.join()

        logging.debug('Waiting for engine workers to finish...')
        self.pool.running = False
        self.pool.join()

        return True


    '''
    begin fuzzing and report progress. do this in foreground
    '''
    def start(self, threads = 4):
        self.solutions = list()

        logging.debug('Initializing engine pool with ', colored_command(threads), ' threads...')
        self.pool = meissner.EnginePool(self, threads = threads)

        logging.debug('Populating jobs list...')
        max_jobs = 100000 # XXX: make dynamic limit
        self.tested = list()
        self.mutator = meissner.Mutator(self, max_jobs)
        self.mutator.start()

        # XXX: this is kinda ugly. just wait until it's not 0
        while len(self.mutator.jobs) == 0:
            time.sleep(1)

        logging.debug('Finished. Starting all engines...')
        self.pool.start()

        try:
            last_tested = 0
            status_wait = 30 # in seconds
            while self.pool.running:
                time.sleep(status_wait)
                rate = ((len(self.tested) - last_tested)/status_wait)

                # give some status information
                info = {
                    'jobs/sec' : '%.2f' % rate + (
                        # https://github.com/Arinerron/meissner/wiki/My-fuzzer-is-running-slowly!
                        # XXX: how slow is "slow"?
                        '\n' if rate > 10 else colored(' (slow? read ' + colored_command('https://bit.ly/39Gggy8') + ')', ['red', 'bold']) + '\n'
                    ),

                    'solutions' : '%d / %d' % (len(self.solutions), len(self.tested))
                }

                max_length = max([len(x) for x in info.keys()])
                message = ['Status update on ', colored_command(time.strftime("%c")), '\n']
                for key, val in info.items():
                    message += [
                        '... ',
                        colored(colored_command(key) + ' '*(max_length-len(key)), ['green', 'bold']),
                        ' | ',
                        colored(val, ['green'])
                    ]
                logging.info(*message)

                last_tested = len(self.tested)

            self.mutator.running = self.pool.running
            logging.error('Meissner finished all fuzzing jobs. Please report this.')
        except KeyboardInterrupt:
            logging.info('Stopping Meissner...')
            self.stop()

        return self.solutions



#!/usr/bin/env python3

import logging
import traceback
import sys, os, io, selectors
import re


# from: https://stackoverflow.com/a/38662876
def strip_ansi(line):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', line)


# cool custom color logging


class ColorLog(object):
    def __init__(self, logger):
        self._log = logger


    def _log_msg(self, name, *args, **kwargs):
        return getattr(self._log, {
            'success' : 'info',
            'exception' : 'fatal'
        }.get(name, name))(self._format_msg(name, *args, **kwargs))


    def _format_msg(self, name, *args, **kwargs):
        exc_info = kwargs.get('exc_info', False)

        if name == 'exception':
            name = 'fatal'
            exc_info = True
        elif name == 'warn':
            name = 'warning'
        elif name == 'critical':
            name = 'fatal'

        _colored = colored if kwargs.get('use_ansi', True) else lambda s, x : s

        prompt = prompts[name] + ' ' if not kwargs.get('prompt') else kwargs.get('prompt')
        message = _colored(prompt, colormap[name]) + ''.join([_colored(x, colormap[name]) for x in args])

        if exc_info:
            exception = traceback.format_exc().strip()

            prompt = '... ' if not kwargs.get('prompt') else kwargs.get('prompt')
            message += '\n' + '\n'.join([
                prompt + _colored(x, colormap['exception']) for x in exception.split('\n')
            ])

        return message


    def __getattr__(self, name):
        if name in ['debug', 'info', 'warn', 'warning', 'error', 'critical', 'fatal', 'exception', 'success']:
            return lambda *args, **kwargs : self._log_msg(name, *args, **kwargs)

        return getattr(self._log, name)



colored = lambda message, attrs = [] : colors['reset'] + ''.join([colors[x] for x in attrs]) + str(message) + colors['reset']
colored_command = lambda message : colors['bold'] + colors['underline'] + str(message) + colors['reset']

colormap = {
    'debug' : ['italics', 'gray'],
    'info' : ['blue'],
    'warning' : ['bold', 'darkorange'],
    'error' : ['bold', 'lightred'],
    'fatal' : ['bg_red', 'bold_white'],
    'exception' : ['italics', 'darkred'],
    'success' : ['bold', 'green']
}

prompts = {
    'debug' : ' * ',
    'info' : '[*]',
    'warn' : '[!]',
    'warning' : '[!]',
    'error' : '[-]',
    'fatal' : '[-]',
    'success' : '[+]'
}

log_levels = {
    'debug' : logging.DEBUG,
    'info' : logging.INFO,
    'warn' : logging.WARNING,
    'warning' : logging.WARNING,
    'error' : logging.ERROR,
    'critical' : logging.CRITICAL,
    'fatal' : logging.FATAL,
    'success' : logging.FATAL,
}

colors = {
    'lightgray' : '\033[37m',
    'darkgray' : '\033[90m',
    'gray' : '\033[2m',
    'blue' : '\033[34m',
    'green' : '\033[32m',
    'cyan' : '\033[36m',
    'darkorange' : '\033[33m',
    'darkred' : '\033[31m',
    'lightred' : '\033[91m',
    'red' : '\033[91m',
    'yellow' : '\033[33m',
    'lightyellow' : '\033[93m',
    'lightgreen' : '\033[92m',
    'bold_white' : '\033[1;37m',

    'bg_red' : '\033[41m',

    'italics' : '\033[3m',
    'bold' : '\033[01m',
    'underline' : '\033[04m',

    'reset' : '\033[0m'
}


LOG_LEVEL = logging.DEBUG


log = ColorLog(logging.getLogger(__name__))
log.setLevel(LOG_LEVEL)
stdout = logging.StreamHandler()
log.addHandler(stdout)
stdout.setLevel(LOG_LEVEL)
logging = log # XXX: find a cleaner solution plz



from . import Harness, placeholder
from meissner.logger import *

import requests, urllib.parse

_default_method = 'GET'

class URLHarness(Harness):
    def __init__(self, url, method = _default_method):

        self.url = url

        # TODO: support POST requests
        self.method = {
            'GET' : requests.get
        }.get(method.strip().upper())

        logging.debug('... using ', colored_command(method.strip().upper()), ' request to URL: ', colored_command(url))

        if not self.method:
            logging.warning('Request method ', colored_command(method), ' is unknown. Defaulting to ', colored_command(_default_method), '...')
            self.method = _default_method


    def test(self, payload):
        url = self.url.format(**{
            placeholder : urllib.parse.quote_from_bytes(bytes(payload), safe = '')
        })

        response = self.method(url)
        return response.content



from meissner.logger import *
from . import Engine, Result
from smartbytes import *

from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

import uuid, base64, time


timeout = 2 # XXX: make this a config!

# how many seconds to wait (e.g. for setTimeout-based XSS) before checking status
wait = 1 # XXX: move to config or find better solution

class SeleniumEngine(Engine):
    def __init__(self):
        self.browser = self._get_browser()

    def _get_browser(self):
        return None

    def test(self, payload):
        # build the test
        # XXX: this makes it so we don't support CSPs
        # XXX: make a custom function instead of alert in case the challenge requires alert function
        indicator = str(uuid.uuid4())
        # XXX: will this work for large pages?
        # XXX: if it loads external resources, we may be screwed
        # TODO: support custom filters like --filter=myfile.py which reads from stdin and writes to stdout
        html = smartbytes('<script>alert=console.log.bind(null,"', indicator, '}")</script>\n\n', payload)
        url = 'data:text/html;base64,' + base64.b64encode(bytes(html)).decode('utf-8')

        # load the html
        self.browser.get(url)

        # XXX: this is vuln to race condition.
        loaded = False
        time.sleep(wait)
        start_time = time.time()
        while time.time() - start_time < timeout:
            time.sleep(0.05)

            # check if page has loaded yet
            if any([
                self.browser.execute_script('return document.readyState;') == 'complete',
                self.browser.current_url != url
                ]):
                loaded = True
                break

        if not loaded:
            logging.warning('Page appears to not have loaded despite being given ', timeout, ' seconds!')

        # check if we won or not
        success = False
        for entry in self.browser.get_log('browser'):
            if indicator in entry.get('message', ''):
                success = True
                break

        result = Result(payload, success)
        return result


class Chrome(SeleniumEngine):
    def _get_browser(self):
        # make it headless
        options = webdriver.ChromeOptions()
        options.add_argument('headless')

        # capture logging
        capabilities = DesiredCapabilities.CHROME
        capabilities['goog:loggingPrefs'] = {
            'browser':'ALL'
        }

        return webdriver.Chrome(chrome_options = options, desired_capabilities = capabilities)


class Firefox(SeleniumEngine):
    def _get_browser(self):
        # make it headles
        os.environ['MOZ_HEADLESS'] = '1'

        # XXX: capture logging?

        return webdriver.Firefox()



from meissner.logger import *

from smartbytes import *
import threading, random, time


class Job(smartbytes):
    # XXX: get rid of this function lol
    def format(self):
        return self


class Mutator(threading.Thread):
    def __init__(self, fuzzer, max_jobs):
        super(Mutator, self).__init__()

        self.meissner = fuzzer
        self.max_jobs = max_jobs
        self.jobs = list()

        self.running = False

    '''
    figure out how to handle various results

    arg job: the Job object representing the payload that was sent
    arg formatted_job: the str representation of the payload physically executed
    arg result: the output from the execution engine
    '''
    def report(self, job, result):
        # XXX: handle other things? like browser-related errors etc
        payload = str(job)
        if result.success and payload not in self.meissner.solutions:
            # XXX: store in file or something?
            border = colored('-'*40, ['blue', 'bold'])
            border2 = colored('-'*40, ['gray']) + '\n'

            # there may be cases where the backend messes with out payload. There is no way to objectively detect this.
            code_highlighter = lambda code : colored(code, ['lightgreen'])
            payload_highlighter = lambda code, x = [] : colored(code, ['red'] + x)

            output = str(result.payload)
            output_split = output.split(payload)

            message = ['Found a functional XSS payload:\n', border, '\n', payload_highlighter(payload), '\n']

            # don't display injected HTML if it is too long
            if len(output) < 5000: # XXX: how do we choose how long of output code is too long
                message.extend([border2, code_highlighter(output_split.pop(0)), '\n'])
                for separator in output_split:
                    message.append(payload_highlighter(payload, ['underline']))
                    message.append(code_highlighter(separator))
            message.append(border)
            message.append('\n')

            logging.success(*message)
            self.meissner.solutions.append(payload)

        # XXX: this is a bad way to mix things up. we just don't want it to get too boring...
        if random.randint(0, 1000) == 0:
            random.shuffle(self.jobs)

        # XXX: determine if this is an interesting case or not?
        self.generate(base = (job, result))


    def mutate(self, possible_job):
        # XXX: once https://github.com/Arinerron/smartbytes/issues/1 is added, use that
        possible_job = list(possible_job)

        # XXX: improve mutation so it's not this sad
        # TODO: this is just temporary so I can get a working PoC, don't judge plz
        for i in range(random.randint(1, 2)):
            possible_job[random.randint(0, len(possible_job)-1)] = random.randint(0x00, 0xff)

        return smartbytes(possible_job)


    '''
    generate jobs dynamically based
    '''
    def generate(self, base = None):
        dictionary = self.meissner.dictionary.copy() + [
            # the empty value is to try only mutating the input without adding anything
            # theoretically all possible exploits will be generated because of this
            '',
            '\n',

            # we also want to increase the chances of the payload having an alert(1) in it
            # XXX: find a better solution
            'alert(1);',
            'alert(1);',
            'alert(1)',
            '*/alert(1);/*',
            'alert(1);//',
        ]
        random.shuffle(dictionary)

        possible_jobs = list()

        if base:
            base_job, result = base
            base_formatted_job = base_job.format()
        else:
            base_job, result = None, None
            base_formatted_job = smartbytes()

        # XXX: improve algorithm. this kinda sucks
        # we shuffle so that if we are at the max jobs size, it
        # gets a chance to "have other payloads" be generated
        for key in dictionary:
            # try just adding it plain
            possible_job = smartbytes(base_formatted_job) + key
            possible_jobs.append(possible_job)

            # and reverse
            possible_job = smartbytes(key) + smartbytes(base_formatted_job)
            possible_jobs.append(possible_job)

            # XXX: mutation is very poorly written and is buggy right now. :(
            # mutate it a little bit if it didn't error
            # if len(possible_job) and result and (result.success or not result.error) and random.randint(0, 10) != 0: # XXX: this is a bad way to throw in randomness
            #     possible_jobs.append(self.mutate(possible_job))

        for job in possible_jobs:
            if len(self.jobs) > self.max_jobs:
                break
            elif job in self.meissner.tested:
                # we've tested this one already :(
                continue

            if 'alert' not in job:
                continue
            # XXX: move mutations down here. also, find a better solution than what's above

            self.jobs.append(Job(job))


    def run(self):
        self.running = True

        self.generate()
        while self.running:
            # busy wait
            if len(self.jobs) > self.max_jobs:
                time.sleep(0.1) # XXX: smarter solution please



from meissner.logger import *


'''
this should allow us to do more complex analysis of
results in the future
'''
class Result:
    def __init__(self, payload, success):
        self.payload = payload
        self.success = success

        # set these by hand for more info
        self.error = False


class Engine:
    def test(self, payload):
        logging.warning('Using dummy engine ', colored_command(self.__class__.__name__))
        return False

from .seleniumengine import SeleniumEngine, Chrome, Firefox



#!/usr/bin/env python

import setuptools
from setuptools import find_namespace_packages
from distutils.core import setup

def readme():
    with open('README.md', 'r') as f:
        return f.read()

setup(name = 'meissner',
      version = '1.0.2',
      description = 'a dictionary-based XSS mutation fuzzer',
      long_description = readme(),
      long_description_content_type="text/markdown",
      author = 'Aaron Esau',
      author_email = 'python@aaronesau.com',
      url = 'https://github.com/Arinerron/meissner',
      keywords = 'ctf xss hacking mutation fuzzer fuzzing fuzz dictionary attack attacking injection scripting',
      packages = ['meissner'] + find_namespace_packages(include=['meissner.*']),
      scripts = [
        'scripts/meissner'
      ],
      install_requires=open('requirements.txt', 'r').read().strip().split('\n'),
      package_data={'meissner': ['dictionary/*.txt']},
      include_package_data=True,
      zip_safe=False,
      python_requires='>=3.6'),



from meissner.logger import *
from meissner.engines import SeleniumEngine

from smartbytes import *
import threading, time


class EngineWorker(threading.Thread):
    def __init__(self, pool, id = 0):
        super(EngineWorker, self).__init__()

        self.pool = pool
        self.id = id
        self._browser_lock = False

        # XXX: do we need a new one every run?
        self.engine = self.pool.meissner.engine()


    def execute(self, job):
        return self.engine.test(str(job))


    def run(self):
        while self.pool.running:
            # busy wait
            if not self.pool.meissner.mutator.jobs:
                time.sleep(0.1)
                continue

            try:
                # get the job OBJECT
                job = self.pool.meissner.mutator.jobs.pop(0)
                self.pool.meissner.tested.append(smartbytes(job))

                # turn it into a str payload
                formatted_job = job.format()

                # pass it through filters
                formatted_job = self.pool.meissner._format(job)

                self._browser_lock = True
                self.pool.meissner.mutator.report(job, self.execute(formatted_job))
                self._browser_lock = False
            except:
                if self.pool.running:
                    logging.exception('Worker ', colored_command(self.id), ' encountered a fatal exception.')
                break


'''
engine pooling
'''
class EnginePool(threading.Thread):
    def __init__(self, fuzzer, threads = 4):
        super(EnginePool, self).__init__()

        self.meissner = fuzzer
        self.threads = list()

        for i in range(1, threads + 1):
            self.threads.append(EngineWorker(self, id = i))

            if i % 10 == 0:
                logging.debug('... created ', colored_command(i), ' workers so far')

        self.running = False


    def run(self):
        self.running = True

        for thread in self.threads:
            thread.start()

        # block until all worker threads are done
        nthreads = len(self.threads)
        i = 0
        while True:
            thread = self.threads[i % len(self.threads)]

            if thread.isAlive() and isinstance(thread.engine, SeleniumEngine):
                # swap out its browser object
                oldbrowser = thread.engine.browser
                engine = self.meissner.engine() # XXX: change this so that we can accept non-selenium engines

                # XXX: race condition :(
                while thread._browser_lock and thread.isAlive():
                    # busy wait
                    time.sleep(0.1)

                thread.engine = engine
                oldbrowser.quit()

                time.sleep(120 / len(self.threads))
            else:
                # otherwise, let's notify that it's done
                thread.join()
                logging.debug('Worker ', colored_command(thread.id), ' finished executing.')

            i += 1
            nthreads -= 1
            if nthreads == 0:
                break



from . import Harness, placeholder
from meissner.logger import *

from smartbytes import *

import subprocess

class CommandHarness(Harness):
    def __init__(self, cmd):
        logging.debug('... using command: ', colored_command(cmd))

        self.cmd = cmd
        self.warn_stderr = True

    def test(self, payload):
        command = self.cmd
        write = None

        if placeholder in command:
            # placeholder was found in command, add as arg
            command = self.cmd.copy()
            i = self.cmd.index(placeholder)
            command[i] = str(payload)
        else:
            # otherwise, write to stdin
            write = bytes(payload)

        p = subprocess.Popen(self, command, stdin = subprocess.PIPE, stdout = subprocess.PIPE, stderr = subprocess.PIPE)

        if write:
            p.stdin.write(payload)

        stdout, stderr = p.communicate()

        # XXX: find a better solution than warn_stderr to not spam console
        if stderr and self.warn_stderr:
            logging.warning('A test case caused data to be written to stderr.\n    case: ', colored_command(str(payload)), '\n    output: ', colored_command(str(smartbytes(stderr))))
            self.warn_stderr = False

        return stdout
